/**
 * Copyright (c) 2016 - 2019, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/** @file
 *
 * @defgroup bootloader_secure_ble main.c
 * @{
 * @ingroup dfu_bootloader_api
 * @brief Bootloader project main file for secure DFU.
 *
 */

#include <stdint.h>
#include "boards.h"
#include "nrf_mbr.h"
#include "nrf_bootloader_info.h"
#include "nrf_bootloader.h"
#include "nrf_bootloader_app_start.h"
#include "nrf_bootloader_dfu_timers.h"
#include "nrf_dfu.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "app_error.h"
#include "app_error_weak.h"
#include "nrf_delay.h"
#include "SEGGER_RTT.h"
#include "nrf_gpio.h"


static void on_error(void)
{
    NRF_LOG_FINAL_FLUSH();

#if NRF_MODULE_ENABLED(NRF_LOG_BACKEND_RTT)
    // To allow the buffer to be flushed by the host.
    nrf_delay_ms(100);
#endif
#ifdef NRF_DFU_DEBUG_VERSION
    NRF_BREAKPOINT_COND;
#endif
    NVIC_SystemReset();
}


void app_error_handler(uint32_t error_code, uint32_t line_num, const uint8_t * p_file_name)
{
	SEGGER_RTT_printf(0, "%s:%d\n", p_file_name, line_num);
    on_error();
}


void app_error_fault_handler(uint32_t id, uint32_t pc, uint32_t info)
{
	SEGGER_RTT_printf(0, "Received a fault! id: 0x%08x, pc: 0x%08x, info: 0x%08x\n", id, pc, info);
    on_error();
}


void app_error_handler_bare(uint32_t error_code)
{
	SEGGER_RTT_printf(0, "Received an error: 0x%08x!\n", error_code);
    on_error();
}

/**
 * @brief Function notifies certain events in DFU process.
 */
static void dfu_observer(nrf_dfu_evt_type_t evt_type)
{
    switch (evt_type)
    {
    	case NRF_DFU_EVT_DFU_INITIALIZED:
    		SEGGER_RTT_printf(0, "%s: NRF_DFU_EVT_DFU_INITIALIZED\n", __func__);
        break;
        case NRF_DFU_EVT_TRANSPORT_ACTIVATED:
        	SEGGER_RTT_printf(0, "%s: NRF_DFU_EVT_TRANSPORT_ACTIVATED\n", __func__);
            break;
        case NRF_DFU_EVT_TRANSPORT_DEACTIVATED:
        	SEGGER_RTT_printf(0, "%s: NRF_DFU_EVT_TRANSPORT_DEACTIVATED\n", __func__);
            break;
        case NRF_DFU_EVT_DFU_STARTED:
            SEGGER_RTT_printf(0, "%s: NRF_DFU_EVT_DFU_STARTED\n", __func__);
            break;
        case NRF_DFU_EVT_OBJECT_RECEIVED:
            SEGGER_RTT_printf(0, "%s: NRF_DFU_EVT_OBJECT_RECEIVED\n", __func__);
            break;
        case NRF_DFU_EVT_DFU_FAILED:
        	SEGGER_RTT_printf(0, "%s: NRF_DFU_EVT_DFU_FAILED\n", __func__);
            break;
        case NRF_DFU_EVT_DFU_COMPLETED:
            SEGGER_RTT_printf(0, "%s: NRF_DFU_EVT_DFU_COMPLETED\n", __func__);
            break;
        case NRF_DFU_EVT_DFU_ABORTED:
        	SEGGER_RTT_printf(0, "%s: NRF_DFU_EVT_DFU_ABORTED\n", __func__);
            break;
        default:
        	SEGGER_RTT_printf(0, "%s: Received event %d\n", __func__, evt_type);
            break;
    }
}

#define UART0_GPIO_PIN		NRF_GPIO_PIN_MAP(1, 0)		// Bootloader Active Signal to host

#define BOOTLOADER_VERSION	"1.0.0.0"

#define SDFIRMWAREIDADDRESS	0x300C

/**@brief Function for application main entry. */
int main(void)
{
    uint32_t ret_val;


    nrf_gpio_cfg(UART0_GPIO_PIN, NRF_GPIO_PIN_DIR_OUTPUT, NRF_GPIO_PIN_INPUT_DISCONNECT, NRF_GPIO_PIN_PULLDOWN, NRF_GPIO_PIN_S0S1, NRF_GPIO_PIN_NOSENSE);
    // Indicate to host that bootloader is running
	nrf_gpio_pin_write(UART0_GPIO_PIN, true);


    SEGGER_RTT_Init();


    uint32_t *pSDFwId = (uint32_t*)SDFIRMWAREIDADDRESS;
	SEGGER_RTT_printf(0, "%s: SoftDevice Firmware Id: 0x%X\n", __func__, (*pSDFwId) & 0xffff);

    // Must happen before flash protection is applied, since it edits a protected page.
    nrf_bootloader_mbr_addrs_populate();

    // Protect MBR and bootloader code from being overwritten.
    ret_val = nrf_bootloader_flash_protect(0, MBR_SIZE, false);
    APP_ERROR_CHECK(ret_val);
    ret_val = nrf_bootloader_flash_protect(BOOTLOADER_START_ADDR, BOOTLOADER_SIZE, false);
    APP_ERROR_CHECK(ret_val);

    (void) NRF_LOG_INIT(nrf_bootloader_dfu_timer_counter_get);
//    NRF_LOG_DEFAULT_BACKENDS_INIT();

    SEGGER_RTT_printf(0, "PR Bootloader %s\n", BOOTLOADER_VERSION);
    SEGGER_RTT_printf(0, "Bootloader start address: 0x%X\n", BOOTLOADER_START_ADDR);

    ret_val = nrf_bootloader_init(dfu_observer);
    APP_ERROR_CHECK(ret_val);

//    NRF_LOG_FLUSH();

    SEGGER_RTT_printf(0, "After main, should never be reached.\n");
//    NRF_LOG_FLUSH();

    APP_ERROR_CHECK_BOOL(false);
}

/**
 * @}
 */
